clear;
clc;

addpath('C:\Code\Matlab\Casadi\casadi-windows-matlabR2016a-v3.5.5')
addpath('C:\Code\Matlab\Master\Spring')
import casadi.*

%% INFO
% Uses "ODE_solver.m" and "NMPC.m"

%% Setup system
% T : Total simulation time
% N : Total number of steps with stepsize dt
% np: Prediction horizon
% nm: Control horizon
% Q: Weighting matrice for setpoint
% R: Weight for input change

% Parameters
system = system_param();
dt = system.dt;                N = system.N;
mu_max = system.mu_max;  Ks = system.Ks;                kd = system.kd;
Y_XS = system.Y_XS ;     Y_CO2X = system.Y_CO2X ;       Sin  = system.Sin;
q_CO2 = system.q_CO2;    pump_speed = system.pump_speed ;

% MPC
system.np = 20;% 20 % 6; % A time horizon of 120 min gives more chattering and longer runtime.
system.nm = 4;%2;
system.Q = 500*[0 0 0 0;
            0 1 0 0;
            0 0 0 0; 
            0 0 0 0];
system.R = 100*0.5;

% Simulation time
T = 24;
N = T/dt;

X_goal = 15; % Biomass setpoint
system.X_goal = X_goal;

% Allocate space for results
tm = zeros(N,1);    % time array
xs = ones(N,4);     % states array 
us = zeros(N,1);    % input array 
cost = zeros(N,1);  % optimization cost
success_rate = zeros(N,1); % log if the solver works

% SP trajectory
x_SP = X_goal*ones(1,N);

%% Variables & Dynamics
% States
V   = SX.sym('V');
X   = SX.sym('X');
S   = SX.sym('S');
CO2 = SX.sym('CO2');
x   = [V; X ; S; CO2];

system.x = x;

% Input
Fin = SX.sym('Fin');
Fout = Fin;
system.u  = [Fin];  

% Differential equations
dV = Fin - Fout;
dX = -(X/V)*Fin + mu_max *(S /(Ks + S)) *X - kd *X ; %Biomass
dS = (Sin *Fin) / V - mu_max *(S /(Ks + S)) *X /Y_XS - (S / V) *Fin; % Substrate
dCO2 = (mu_max *(S /(Ks + S)) *X)/Y_CO2X - q_CO2*CO2 ;

xdot = [dV, dX, dS, dCO2]';
system.xdot = xdot;

%% Initial condiitons
x0 = [1.5, 0.8, 15, 0]'; % 1.0, 1.2, 20, 0
xk = x0;
xs(1,:) = xk;

t0 = tm(1);
u0 = us(1); % Should be = 0
uk = u0;

%% Simulation
system.noise = 0;  % ! Must be included for ODE solver to run
% CALL PLANT / solver
[Plant] = ODE_solver(system);

%% Run MPC

% TEST NMPC
%{
xk_SP_traj = x_SP(1,1:end);
u_opt = NMPC(xk, uk, xk_SP_traj, system);


x1_opt = u_opt(1:5:end);
x2_opt = u_opt(2:5:end);
x3_opt = u_opt(3:5:end);
x4_opt = u_opt(4:5:end);
u_opt_ = u_opt(5:5:end);

tgrid = linspace(0, system.np, system.np+1);

clf;
figure(1)
plot(tgrid, x1_opt)
figure(2)
hold on
plot(tgrid, x2_opt, 'r')
yline(xk_SP_traj(1))
figure(3)
plot(tgrid, x3_opt, 'r')
figure(4)
plot(tgrid, x4_opt, 'r')
figure(5)
stairs(tgrid, [u_opt_',nan], '-')
%stairs(tgrid, [y_SP nan], '-')

%legend('x1','x2', 'x2SP','x3', 'x4', 'u')

%}

tic
tStart = cputime;
for i = 1:N-1%
    fprintf('>>> Iteration: %d \n',i)
    % Get current and future SP trajectories
    xk_SP_traj = x_SP(:,i:end);
    
    % Calculate optimal input (MV) for nm, simulate CV for np
    [u_opt, w_cost, success] = NMPC(xk, uk, xk_SP_traj, system, i);
    
    % Update state & implement control
    uk = u_opt(1);
    success_rate(i) = success;
    
    % Integrate with optimal input
    sim = Plant('x0', xk, 'p', [uk] );
    xk = full(sim.xf);
    
    %for plotting
    tm(i + 1)   =  i*dt;
    xs(i + 1,:) =  xk;
    us(i)       =  uk;
    cost(i)     =  w_cost;
    
    
end
tEnd = cputime - tStart
endtime = toc
%% Plotting
% Check which iterations failed
isNZ=(~success_rate==0);           % addressing logical array of nonzero elements

% Get colors
color = colorHex();

clf;

 set(0, 'DefaultTextFontName', 'Computer Modern',...
     'DefaultTextFontSize', 10,...
     'DefaultAxesFontName', 'Computer Modern',...
     'DefaultAxesFontSize', 10,...
     'DefaultLineLineWidth', 1.2)

set(gcf,'Position',[450 300 650 580])

figure(1)
subplot(4,1,1), %plot states
    hold on;
    plot(tm, xs(:,1), 'Color', color.r)

    xlabel('t [h]')
    ylabel('Volume [L]')
    legend boxoff
    legend({'V: Volume'}, 'Location', 'best');%'Position',[0.8 0.9 0.15 0.08]);
    title('Volume')
    

subplot(4,1,2), %plot states
    hold on;
    plot(tm, xs(:,2), 'Color', color.r)
    plot(tm, xs(:,3),'Color',color.p)
    yline(X_goal,'--', 'X_{goal}')
    %ylim([0,X_goal+3])
    xlabel('t [h]')
    ylabel('Concentration [g/L]')
    legend boxoff
    legend({'X: Cell','S: Glucose'},'Location', 'SouthEast')%'Position',[0.8 0.7 0.15 0.08]);
    title('Cell and Sugar')

 subplot(4,1,3), %plot states
    hold on;
    plot(tm, xs(:,4),'Color', color.r)
    
    xlabel('t [h]')
    ylabel('% of outlet')
    legend boxoff
    legend({'CO_2'}, 'Location', 'best')%'Position',[0.8 0.5 0.15 0.08]);
    title('Offgas')

subplot(4,1,4), %plot states
    hold on;
    colororder({'k','k'})
    stairs(tm, us, 'Color', color.r, 'LineWidth',0.9)
    xlabel('t [h]')
    ylabel('Flow [L/h]')
    yyaxis right
    ylabel('Cost')
    plot(tm, cost, '--', 'Color',color.b)
    legend boxoff
    legend('F_{in} = F_{out}', 'Cost', 'Location', 'best')
    title('Feed')
sgtitle('Optimal feeding profile, continuous')

%%Plot feed), %plot states

figure(2)
hold on
colororder({'k','k'})
stairs(tm, us, 'Color', color.r, 'LineWidth',1.5)
xlabel('t [h]')
ylabel('L/h')
yyaxis right
ylabel('Cost')
%plot(tm(isNZ),success_rate(isNZ)) % plot failed iterations
plot(tm, cost, '--', 'Color',color.b)
%legend('Fin = Fout', 'Failed optimization', 'Cost', 'Location', 'best')%'Position',[0.8 0.25 0.15 0.08]);
legend('Fin = Fout','Cost', 'Location', 'best')
legend boxoff
title('Optimal feeding profile, Continous')

%% Edit profile 
% shift to start after 6 hours - as this is experienced batch phase. 
u_save = us;
u_first = find( us >0, 1, 'first');
u_end = find( us >0, 1, 'last');

u_range = u_end - u_first;

u_end_val = us(u_end );

t_shift_feed = find( tm <= 6 , 1, 'last');
% shift
us_shift = us(u_first:u_end);
us(t_shift_feed:t_shift_feed+u_range) = us_shift;
us(t_shift_feed+u_range:end) = u_end_val;


fig = figure; clf
set(gcf,'Position',[450 300 550 200])
hold on
stairs(tm, us, 'Color', color.r, 'LineWidth',1.5)
xlabel('Time [h]')
ylabel('L/h')
colororder({'k','k'})
stairs(tm, u_save, '.-','Color', color.p, 'LineWidth',1.5)
yyaxis right
ylabel('Cost')
%plot(tm(isNZ),success_rate(isNZ)) % plot failed iterations
plot(tm, cost, '--', 'Color',color.b)
legend('Shifted: F_{in} = F_{out}','F_{in} = F_{out}', 'Cost', 'Location', 'best')
legend boxoff
title('Optimal flow, continous cultivation')



%% PLOT Both states
x_shifted = zeros(N,4);
x_shifted(1,:) = x0;

xk = x0;

for i=1:N-1
    uk = us(i);
    sim = Plant('x0', xk, 'p', [uk] );
    xk = full(sim.xf);
    x_shifted(i+1,:) = xk;
end



 set(0, 'DefaultTextFontName', 'Computer Modern',...
     'DefaultTextFontSize', 10,...
     'DefaultAxesFontName', 'Computer Modern',...
     'DefaultAxesFontSize', 10,...
     'DefaultLineLineWidth', 1.2)

set(gcf,'Position',[450 300 650 580])


tile  = tiledlayout(2,2,'TileSpacing','Compact','Padding','Compact');

%% Volume
nexttile
    hold on
    
    plot(tm,    x_shifted(:,1),   '-','Color', color.r)
    plot(tm, xs(:,1),'Color', color.p)

    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('V [L]')
    %legend('boxoff')
   % legend('Model','MHE','Measurement','location', 'best')

%% Biomass
nexttile
    hold on
    plot(tm,    x_shifted(:,2),   '-','Color', color.r)
    plot(tm, xs(:,2), 'Color', color.p)
    yline(X_goal,'--', 'X_{goal}')
    axis tight
    xlabel('Time [h]')
    ylabel('X [g/L]')
    legend('boxoff')
    legend('Shifted Input', 'Original','location', 'best')

%% Sugar
nexttile
    hold on
    
    plot(tm,    x_shifted(:,3),   '-','Color', color.r)
    plot(tm, xs(:,3), '-','Color', color.p)
    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('S [g/L]')
    %legend('boxoff')
    %legend('Model','MHE','Offline','location', 'best')    %'off-line S',

%% CO2
nexttile
    hold on
    
    plot(tm,    x_shifted(:,4),   '-','Color', color.r)
    plot(tm, xs(:,4), '-','Color', color.p)
    % add axis labes and legend
    axis tight
    xlabel('Time [h]')
    ylabel('CO2 [% of gas outlet]')
    %legend('boxoff')
    %legend('Model','MHE','Measurement','location', 'best')
    
sgtitle("Dynamics, Optimal Input",  'fontweight','bold')

%% Save results
V = xs(:,1); X =xs(:,2); S = xs(:,3); CO2 = xs(:,4); Fin = us;
% Create table with data
time = tm ;
T_feed_profile = table(time, V, X, S, CO2, Fin);

col_header={'timeSim','V','X','S','CO2','Fin'};

%Write to file
filename_feed_profile = 'ideal_feed_profile_param_est_umin_08.xlsx';

writetable(T_feed_profile,filename_feed_profile, 'Sheet',1, 'WriteVariableNames',true)
