    function [u_opt, w_cost, success]  = NMPC(x0, u0,xk_SP_traj,system,i)
    import casadi.*
    %% System model
    % Control parameters
    np = system.np ;         % prediction horizon
    nm = system.nm ;         % control horizon 
    
    % Tuning parameters
    Q = system.Q;
    R = system.R ;
    
    % Time step
    dt = system.dt;

    % Parameter values
    mu_max = system.mu_max;  Ks = system.Ks;                kd = system.kd;
    Y_XS = system.Y_XS ;     Y_CO2X = system.Y_CO2X ;       Sin  = system.Sin;
    q_CO2 = system.q_CO2;    
    
    pump_speed = system.pump_speed;

    %% Bounds
    % Force continous production after feeding has been initiated
    if i > 800
        umin = 0.8/45.0196; % 0.8% is the lowest effective pump rate. 
    else
        umin = 0*pump_speed;
    end
    % Max flow
    umax = 100*pump_speed; % 100 percent pump flow
    
    % Bounds in flow changes
    du_max = 50*pump_speed; %0.00210; % L/h
    du_max_= 80*pump_speed;

    %Vmin = 1.0;
    Vmax = 2.5; % Max 2.5 Liters

    X_goal = system.X_goal; % g/L

    %% Trajectory
    % Avoid iterating outside array at the end of the time horizon
    if length(xk_SP_traj) < np
        np = length(xk_SP_traj);
        nm = np;
    end
    %% Declare model variables
    x = system.x;   % states
    num_states = length(x);
    
    u = system.u;   
    
    %% System differential equations
    xdot = system.xdot; % Differential equations
    
    %% Multiple shooting
    % Objective function
    L =0 ;
    
    % Discrete time dynamics - allows for evaluation at every point
    dae = struct('x',x,'p',u,'ode',xdot,'quad',L);
    % building the integrator
    opts = struct('tf',dt); 
    F = integrator('F', 'cvodes', dae, opts);

    % Optimization problem
    % preparing symbolic variables for NLP
    w = {};

    % preparing numeric variables and bounds
    w0  = [];
    lbw = [];
    ubw = [];

    % preparing symbolic constraints
    g   = {};
    % preparing numeric bounds
    lbg = [];
    ubg = [];

    % Prepare cost function aggregation
    J   = 0;

    %% Initial state
    Xk = MX.sym('X0', num_states);
    
    % ! x0 = current xk in MPC outer loop
    w = {w{:}, Xk};
    lbw = [lbw; x0(1); x0(2); x0(3); x0(4)];
    ubw = [ubw; x0(1); x0(2); x0(3); x0(4)];
    w0 =  [w0;  x0(1); x0(2); x0(3); x0(4)];

    uk = u0;
    Uks = [];
    Uks = [Uks, uk]; % Current input

    %% Integrating
    % timestep is still dt (ensured in F) -> this just keeps track of which
    % horizon step we are on -> 1,2,...,np
    step = 1;
    for i = 1:step:np %np

         % Setting up U
         Uk = MX.sym(['U_' num2str(i)]);
         Uks = [Uks, Uk];
         % declaring symbolic variables
         w = {w{:}, Uk};
         
         % declaring them numerically
         w0 = [w0; uk]; %initial guess ; D0
         % bounds
         lbw = [lbw; umin]; 
         ubw = [ubw; umax];

         % State integration for interval
         Fk = F('x0',Xk,'p',Uk);

         Xk_end = Fk.xf;
            
         % Define NEW state variable for the end of the interval
         Xk = MX.sym(['X_' num2str(i+1)], num_states);
         w = [w, {Xk}];
         
         lbw = [lbw;     0;       0;         0;      0];
         ubw = [ubw;  Vmax;     inf;       inf;    inf];
         w0  = [w0;  x0(1);   x0(2);     x0(3);  x0(4)]; 
    
         % Declaring constraints -> Shooting gap
         g   =  [g, {Xk_end - Xk}];
         
         lbg = [lbg; 0; 0; 0; 0]; %  xk_end - xk = 0
         ubg = [ubg; 0; 0; 0; 0];

         % REFERENCE STATE
         xk_SP = xk_SP_traj(i); 

         % Cost function -> penalty offset from SP
         % Will only include Xk(1) as Q only has value for this point
         J = J + 0.5* ( (Xk - xk_SP)'*Q*(Xk - xk_SP)) ;               %+ Fk.qf; 

         % Cost function -> penalty large change in u
         if i > nm
             % No change in u in t in [nm, np]
             g  = [g, {Uk-Uks(i)}];
            lbg = [lbg; 0]; 
            ubg = [ubg; 0];
         else
             % Only add cost until nm
             J  = J + 0.5*R*(Uk-Uks(i))^2;
             g  = [g, {Uk-Uks(i)}];
            lbg = [lbg;    -du_max_]; 
            ubg = [ubg;     du_max];
         end


    end

    % formalize it into an NLP problem
    nlp = struct('x', vertcat(w{:}), 'g', vertcat(g{:}), 'f', J);

    % assign solver - IPOPT in this case
    
    % Options
    options =struct;
    options.ipopt.print_level = 0;  %Remove output from IPOPT
    options.ipopt.tol = 1e-6;
    options.ipopt.acceptable_tol = 1e-2;
    solver = nlpsol('solver', 'ipopt', nlp, options);

    % solve - using the previous defined initial guess and bounds
    sol = solver('x0', w0, 'lbx', lbw, 'ubx', ubw, 'lbg', lbg, 'ubg', ubg); %parameters
    
    % For error messages: Rows correspond to constraints (scalarized) and columns to variables (scalarized). 

    % Handle solver failure
    if solver.stats.success == 0
        disp('WARNING NOT OPTIMZED')
        disp(full(sol.x));
    else
        disp('OPTIMIZED')
    end
    
    success = solver.stats.success;
    
    % Extract optimal solution
    w_sol = full(sol.x);
    w_cost = full(sol.f);
    disp(w_cost);
    
    % Get input solution 
    u_opt_temp = w_sol(5:5:end); % optimal feed every fifth results (remaining are states)
    
    u_opt = [u_opt_temp(1:nm)',u_opt_temp(nm)*ones(1,np-nm)]; % Const u outside control horizon
    
    % For test run of 1 iteration
    % u_opt = full(sol.x);
    % disp(Uks(1))
    % uk = u_opt(5:5:end);
    % disp(uk)
end