

function results = Parameter_estimation(initParams) 
%% This code performs parameter estimation based on the least squares estimator. 
% This is done based on the data points that have sugar measurements
% Initial guess, bounds, and tolerances are tuned to preference

% Get system parameters
addpath('C:\Code\Matlab\Master\Spring')
system = system_param();

% Initial guess
mu_max_0 = 0.22846 ;  Ks_0 = 0.0077687  ;  kd_0 =  0.0025287;  Y_XS_0 = 0.43004  ; Y_CO2X_0 = 0.39328;


%% Get CDW sample data
filename = 'CDW_20230314.xlsx';

lab_result_table = readtable(filename, 'Sheet','Result_14_03');
lab_results = table2array(lab_result_table(:,4:12)); 

dryweight = [lab_results(:,4:6)-lab_results(:,1:3)]; % Triplicates

time_CDW = lab_results(2:7,9); 

cell_density = dryweight(2:7,:)*1000/3; % g/L, mL-> L, 3 mL samples -> x g in 3mL  /3 -> y g/L

mean_CDW = mean(cell_density,2);
std_CDW = std(cell_density,0, 2);

%% Data 
% load experimental data
exp_data = importdata('data_20230314.mat');

meas    = [exp_data.V, exp_data.X, exp_data.S_tab, exp_data.CO2];
ts      = exp_data.ts;    % [s]
Fin      = exp_data.F_in; % [L/h]
Fout     = exp_data.F_out;% [L/h]

t_tot = ts;

%% Filtrer data
%Find nan rows (all rows that do not have data on sugar)
[row, col] = find(isnan(meas));

% Remove nan rows by index => all data is same size
meas( row, : )  = [];
ts( row )       = [];

% Initial state
x0 = meas(1, :);

% Number of datasamples
N = length(ts);

% input u
n_rows = length(t_tot); % want ALL inputs for simulation with new parameters

u = zeros(n_rows,2);
u(:,1) = Fin;
u(:,2) = Fout;

% Define m for residuals
m_max = max(meas(:,2:4)); % scale results for better runtime
m = meas(:,2:4)./m_max;   % volume mismatch not included in square error

% Options for optimizer 
% take large steps to reduce number of iterations & small enough steps to meet a desired accuracy 
opts = optimset('TolFun', 1e-6, 'TolX', 1e-6,'MaxFunEvals', 10e3, 'MaxIter', 10e4, 'FinDiffRelStep',1e-10, 'Diagnostics', 'off', 'Display', 'iter'); % 200 iter

% bounds for parameters we want to estimate
% All params
lb_theta = [0.09; 0.001 ; 0.001; 0.1 ;   0.1]; 
ub_theta = [0.3; 0.01;   0.009;   0.5 ;   0.9];

% initial values for unknown parameters, use if not passed by calling the
% function-handle to this file
initParams = [mu_max_0; Ks_0; kd_0;  Y_XS_0;  Y_CO2X_0]';

% Nonlinear least squares solver, x = lsqnonlin(fun,x0,lb,ub,options)
[theta ,resnorm,res,exitflag,output,lambda,jacobian]= lsqnonlin(@residual,  initParams, lb_theta,ub_theta, opts);

plot_results(theta,ts, meas, time_CDW,mean_CDW,std_CDW)

disp('Parameters: '); disp(theta);

disp('Number of datapoints');  disp(length(ts));

% Find variance through FIM
%Estimated parameter variance
Jacobian = full(jacobian); %lsqnonlin returns the Jacobian as a sparse matrix

var_theta = resnorm*inv(Jacobian'*Jacobian)/N;  % Variance parameters
std_theta = sqrt(diag(var_theta));              % Standard deviation of parameters
std_theta_perc = 100*std_theta'./theta;         % [%]

cov_theta = inv(var_theta);                     % Parameter covariance matrix

disp([' \mu_{max}: ', num2str(theta(1)), ' +/- ', num2str(std_theta(1))])
disp([' K_s: ', num2str(theta(2)), ' +/- ', num2str(std_theta(2))]);
disp([' k_d: ', num2str(theta(3)), ' +/- ', num2str(std_theta(3))]);
disp([' Y_{XS}: ', num2str(theta(4)), ' +/- ', num2str(std_theta(4))]);
disp([' Y_{CO_2}: ', num2str(theta(5)), ' +/- ', num2str(std_theta(5))]);


disp('The parameter covariance matrix')
disp(cov_theta)
a= diag(var_theta);
disp('Variance')
disp(diag(var_theta))
disp(std_theta.^2)

    % See solution of each iteration
   function R = residual(theta_k)
         
       % Solve for time and state according to new parameters (include all original timepoints)
       [~ ,yk ]= solve (theta_k);
       
       % only keep values for corresponding timepoints
       yk( row, : )  = [];
       
       y = yk(:,2:4)./m_max; % volume mismatch not included in square error, centering and scaling helps solve problem
       
       % compute error
       residual = (m(:) - y(:));   % ! solver automatically squares, so no ^2 here !
       R = residual;
       
       results = theta_k;

   end


function [ T, Y ] = solve(theta_k)
  
mu_max  = theta_k(1);
Ks      = theta_k(2);
kd      = theta_k(3);
Y_XS    = theta_k(4);
Y_CO2X  = theta_k(5);

% x0 global
% solve differential for all timesteps
[T,Y] = ode45(@(t,xk) dynamics(t, xk, u, t_tot), t_tot, x0, []);



function dx = dynamics(t, xk, u, time_u)
% find current iteration based on time
u_i = find( t_tot <= t, 1, 'last');

% input for current timestep
Fin_k  = u(u_i,1);
Fout_k = u(u_i,2);

% current state
V   = xk(1);    
X   = xk(2);
S   = xk(3);
CO2 = xk(4);

% fixed parameters
Sin         = system.Sin;         % g/L
q_CO2       = system.q_CO2;       % L/h

% Define differential equations
dV = Fin_k - Fout_k;
dX = -(X/V)*Fin_k + mu_max *(S /(Ks + S)) *X - kd *X ; %Biomass
dS = (Sin *Fin_k) / V - mu_max *(S /(Ks + S)) *X /Y_XS - (S / V) *Fin_k; % Substrate
dCO2 = (mu_max *(S /(Ks + S)) *X)/Y_CO2X - q_CO2*CO2 ;


% Save differential equations in one vector
dx = [dV, dX, dS, dCO2]';

end

end

function plot_results(theta_k,ts, meas, time_CDW,mean_CDW,std_CDW)
    [~ ,yk ]= solve (theta_k);
    
    % Find corresponding CDW data
    t_CDW = zeros(length(time_CDW),1);
    param_CDW = zeros(length(time_CDW),1);

    for i = 1:length(time_CDW)
        i_CDW = find(ts <= time_CDW(i), 1, 'last');
        t_CDW(i) = ts(i_CDW);
        param_CDW(i) = yk(i_CDW,2);
    end
    
    % Calculate RMSE
    % Get data at offline sample timestamps
    yk( row, : )  = [];
    
    MS_CDW = [(param_CDW - mean_CDW).^2  ];
    MS    = [ (meas(:,2) - yk(:,2)).^2  (meas(:,3) -  yk(:,3)).^2];
    N_points = length(meas(:,1));
    RMSE = sqrt(sum(MS)/N_points);
    disp("Root mean squared error:")
    disp(RMSE)
    
    N_CDW = length(mean_CDW);
    RMSE_CDW = sqrt(sum(MS_CDW)/N_CDW);
    disp("Root mean squared error CDW:")
    disp(RMSE_CDW)
    
    % PLOT
    % Display parameters
          formatSpec = ' \x03bc_{max} = %0.5g,  K_s = %0.5g, k_d = %0.5g,  Y_{XS}= %0.5g,  Y_{CO_2X} = %0.5g';
          str = sprintf(formatSpec, theta_k(1),theta_k(2),theta_k(3),theta_k(4), theta_k(5));
          figure(1)
          Title_btm = title(str);
          
          color = colorHex();

        set(0, 'DefaultTextFontName', 'Computer Modern',...
         'DefaultTextFontSize', 10,...
         'DefaultAxesFontName', 'Computer Modern',...
         'DefaultAxesFontSize', 10,...
         'DefaultLineLineWidth', 1)
    
        set(gcf,'Position',[450 200 650 680])
          
          subplot(4,1,1);
          plot(ts,yk(:,1), 'Color', color.b);
          %ylim([0.0 2.0]); % L
          hold on;
          plot(ts,meas(:,1),'+', 'Color', [0.5 0.5 0.5]);
          xlabel('Time [h]')
          ylabel('V [L]')
           title(str);
          hold off;
          
          subplot(4,1,2);
          plot(ts,yk(:,2), 'Color', color.b);
          %ylim([0.0 30.0]); % g/L
          hold on;
          plot(ts,meas(:,2),'+', 'Color', [0.5 0.5 0.5]);
          plot(time_CDW, mean_CDW, 'x', 'Color', color.r)
          errorbar(time_CDW, mean_CDW, std_CDW, 'Color', color.r,'LineStyle', 'none')
          xlabel('Time [h]')
          ylabel('X [g/L]')
          legend('Fitted data', 'Measurement', 'Offline', 'Location', 'northwest')
          legend('boxoff')
          hold off;
         
          subplot(4,1,3);
          plot(ts,yk(:,3), 'Color', color.b);
          %ylim([0.0 50.0]); % g/L
          hold on;
          plot(ts,meas(:,3),'x', 'Color', color.r);
          xlabel('Time [h]')
          ylabel('S [g/L]')
          hold off;
          
          subplot(4,1,4);
          plot(ts,yk(:,4), 'Color', color.b);
          %ylim([0.0 inf]); % g/L
          hold on;
          plot(ts,meas(:,4),'+', 'Color', [0.5 0.5 0.5]);
          xlabel('Time [h]')
          ylabel('CO_2 % of gas outlet')
          
          sgtitle("Parameter Estimation")
         
          hold off;
          
          hold off;
          drawnow;
end

end